################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import six
import os
import numpy as np
import pandas as pd
from datetime import date, datetime
from functools import partial
from types import MethodType

if os.name == 'nt':
    BINDING = 'libbinding.pyd'
    PSP = 'libpsp.dll'
else:
    BINDING = 'libbinding.so'
    PSP = 'libpsp.so'

# rename libbinding.so and libpsp.so temporarily to ensure that client mode
# works automatically when the C++ build fails.
lib_path = os.path.join(os.path.dirname(__file__), "..", "..", "table")
binding = os.path.join(lib_path, BINDING)
psp = os.path.join(lib_path, PSP)
new_binding = os.path.join(lib_path, "notlibbinding.so")
new_psp = os.path.join(lib_path, "notlibpsp.so")


def mock_post(self, msg, msg_id=None, assert_msg=None):
    '''Mock the widget's `post()` method so we can introspect the contents.'''
    assert msg == assert_msg


def setup_module():
    os.rename(binding, new_binding)
    os.rename(psp, new_psp)
    assert os.path.exists(new_binding)
    assert os.path.exists(new_psp)
    assert not os.path.exists(binding)
    assert not os.path.exists(psp)


def teardown_module():
    os.rename(new_binding, binding)
    os.rename(new_psp, psp)
    assert os.path.exists(binding)
    assert os.path.exists(psp)
    assert not os.path.exists(new_binding)
    assert not os.path.exists(new_psp)


class TestClient(object):

    def test_widget_client(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": [i for i in range(50)]}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == data

    def test_widget_client_np(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": [i for i in range(50)]
        }

    def test_widget_client_df(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = pd.DataFrame({
            "a": np.arange(10),
            "b": [True for i in range(10)],
            "c": [str(i) for i in range(10)]
        })
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "index": [i for i in range(10)],
            "a": [i for i in range(10)],
            "b": [True for i in range(10)],
            "c": [str(i) for i in range(10)]
        }

    def test_widget_client_np_structured_array(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = np.array([(1, 2), (3, 4)], dtype=[("a", "int64"), ("b", "int64")])
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": [1, 3],
            "b": [2, 4]
        }

    def test_widget_client_np_recarray(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = np.array([(1, 2), (3, 4)], dtype=[("a", "int64"), ("b", "int64")]).view(np.recarray)
        widget = perspective.PerspectiveWidget(data)
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": [1, 3],
            "b": [2, 4]
        }

    def test_widget_client_schema(self):
        import perspective
        assert perspective.is_libpsp() is False
        widget = perspective.PerspectiveWidget({
            "a": int,
            "b": float,
            "c": bool,
            "d": date,
            "e": datetime,
            "f": str
        })
        assert hasattr(widget, "table") is False
        assert widget._data == {
            "a": "integer",
            "b": "float",
            "c": "boolean",
            "d": "date",
            "e": "datetime",
            "f": "string"
        }

    def test_widget_client_schema_py2_types(self):
        import perspective
        assert perspective.is_libpsp() is False
        if six.PY2:
            widget = perspective.PerspectiveWidget({
                "a": long,  # noqa: F821
                "b": float,
                "c": bool,
                "d": date,
                "e": datetime,
                "f": unicode  # noqa: F821
            })
            assert hasattr(widget, "table") is False
            assert widget._data == {
                "a": "integer",
                "b": "float",
                "c": "boolean",
                "d": "date",
                "e": "datetime",
                "f": "string"
            }

    def test_widget_client_update(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        comparison_data = {
            "a": [i for i in range(50)]
        }
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={
            "cmd": "update",
            "data": comparison_data
        })
        widget.post = MethodType(mocked_post, widget)
        widget.update(data)
        assert hasattr(widget, "table") is False

    def test_widget_client_update_cached(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": [1, 2]}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={
            "cmd": "update",
            "data": data
        })
        widget.post = MethodType(mocked_post, widget)
        widget.update(data)
        assert widget._predisplay_update_cache == [data]
        widget._displayed = True
        widget.update(data)
        assert widget._predisplay_update_cache == [data]

    def test_widget_client_replace(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        new_data = {"a": [1]}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={
            "cmd": "replace",
            "data": new_data
        })
        widget.post = MethodType(mocked_post, widget)
        widget.replace(new_data)
        assert widget._data is new_data

    def test_widget_delete_client(self):
        import perspective
        assert perspective.is_libpsp() is False
        data = {"a": np.arange(0, 50)}
        widget = perspective.PerspectiveWidget(data)
        mocked_post = partial(mock_post, assert_msg={
            "cmd": "delete"
        })
        widget.delete()
        widget.post = MethodType(mocked_post, widget)

    def test_widget_load_column_pivots_client(self):
        import perspective
        assert perspective.is_libpsp() is False
        # behavior should not change for client mode
        arrays = [np.array(['bar', 'bar', 'bar', 'bar', 'baz', 'baz', 'baz', 'baz', 'foo', 'foo', 'foo', 'foo', 'qux', 'qux', 'qux', 'qux']),
                  np.array(['one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two', 'one', 'one', 'two', 'two']),
                  np.array(['X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y', 'X', 'Y'])]
        tuples = list(zip(*arrays))
        index = pd.MultiIndex.from_tuples(tuples, names=['first', 'second', 'third'])
        df_both = pd.DataFrame(np.random.randn(3, 16), index=['A', 'B', 'C'], columns=index)
        widget = perspective.PerspectiveWidget(df_both)
        assert hasattr(widget, "table") is False
        assert widget.columns == [' ']
        assert widget.column_pivots == ['first', 'second', 'third']
        assert widget.row_pivots == ['index']


################################################################################
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#

import os
import pandas
from functools import partial, wraps
from random import random

from .view_config import ViewConfig
from ._data_formatter import to_format, _parse_format_options
from ._constants import COLUMN_SEPARATOR_STRING
from ._utils import _str_to_pythontype
from ._callback_cache import _PerspectiveCallBackCache
from ._date_validator import _PerspectiveDateValidator
from .libbinding import make_view_zero, make_view_one, make_view_two,\
    to_arrow_zero, to_arrow_one, to_arrow_two, get_row_delta_zero,\
    get_row_delta_one, get_row_delta_two


class View(object):
    '''A :class:`~perspective.View` object represents a specific transform
    (pivot, filter, sort, etc) configuration on an underlying
    :class:`~perspective.Table`.  :class:`~perspective.View` objects
    cannot be directly instantiated - they must be derived from an existing
    :class:`~perspective.Table` via the :func:`~perspective.Table.view()`
    method.

    :class:`~perspective.View` instances receive all updates from the
    :class:`~perspective.Table` from which they are derived, and can be
    serialized (via ``to_*`` methods) or trigger a callback when it is updated.
    :class:`~perspective.View` objects will remain in memory and actively
    process updates until :obj:`~perspective.View.delete()` method is called.
    '''

    def __init__(self, Table, **kwargs):
        self._name = "py_" + str(random())
        self._table = Table
        self._config = ViewConfig(**kwargs)
        self._sides = self.sides()

        date_validator = _PerspectiveDateValidator()

        if self._sides == 0:
            self._view = make_view_zero(self._table._table, self._name, COLUMN_SEPARATOR_STRING, self._config, date_validator)
        elif self._sides == 1:
            self._view = make_view_one(self._table._table, self._name, COLUMN_SEPARATOR_STRING, self._config, date_validator)
        else:
            self._view = make_view_two(self._table._table, self._name, COLUMN_SEPARATOR_STRING, self._config, date_validator)

        self._column_only = self._view.is_column_only()
        self._callbacks = self._table._callbacks
        self._delete_callbacks = _PerspectiveCallBackCache()
        self._client_id = None

    def get_config(self):
        '''Returns a copy of the immutable configuration ``kwargs`` from which
        this :class:`~perspective.View` was instantiated.

        Returns:
            :obj:`dict`: ``kwargs`` supplied to the
                :func:`perspective.Table.view()` method.
        '''
        return self._config.get_config()

    def sides(self):
        '''An integer representing the # of hierarchial axis on this
        :class:`~perspective.View`.

        0 - Neither ``row_pivots`` nor ``column_pivots`` properties are set.

        1 - ``row_pivots`` is set.

        2 - ``column_pivots`` is set (and also maybe ``row_pivots``).

        Returns:
            :obj:`int`: 0 <= N <= 2
        '''
        if len(self._config.get_row_pivots()) > 0 or len(self._config.get_column_pivots()) > 0:
            if len(self._config.get_column_pivots()) > 0:
                return 2
            else:
                return 1
        else:
            return 0

    def num_rows(self):
        '''The number of aggregated rows in the :class:`~perspective.View`.

        This count includes the total aggregate rows for all ``row_pivots``
        depth levels, and can also be affected by any applied ``filter``.

        Returns:
            :obj:`int`: Number of rows.
        '''
        return self._view.num_rows()

    def num_columns(self):
        '''The number of aggregated columns in the :class:`~perspective.View`.
        This is affected by the ``column_pivots`` that are applied to the
        :class:`~perspective.View`.

        Returns:
            :obj:`int`: Number of columns.
        '''
        return self._view.num_columns()

    def get_row_expanded(self, idx):
        '''Returns whether row at `idx` is expanded or collapsed.

        Returns:
            :obj:`bool`: Is this row expanded?
        '''
        return self._view.get_row_expanded(idx)

    def expand(self, idx):
        '''Expands the row at 'idx', i.e. displaying its leaf rows.

        Args:
            idx (:obj:`int`): Row index to expand.
        '''
        return self._view.expand(idx, len(self._config.get_row_pivots()))

    def collapse(self, idx):
        '''Collapses the row at 'idx', i.e. hiding its leaf rows.

        Args:
            idx (:obj:`int`): Row index to collapse.
        '''
        return self._view.collapse(idx)

    def set_depth(self, depth):
        '''Sets the expansion depth of the pivot tree.

        Args:
            depth (:obj:`int`): Depth to collapse all nodes to, which
                may be no greater then the length of the ``row_pivots``
                property.
        '''
        return self._view.set_depth(depth, len(self._config.get_row_pivots()))

    def column_paths(self):
        '''Returns the names of the columns as they show in the
        :class:`~perspective.View`, i.e. the hierarchial columns when
        ``column_pivots`` is applied.

        Returns:
            :obj:`list` of :obj`str`: Aggregated column names.
        '''
        paths = self._view.column_paths()
        string_paths = []

        for path in paths:
            string_paths.append(COLUMN_SEPARATOR_STRING.join([p.to_string(False) for p in path]))

        return string_paths

    def schema(self, as_string=False):
        '''The schema of this :class:`~perspective.View`, which is a key-value
        map that contains the column names and their Python data types.

        If the columns are aggregated, their aggregated types will be shown
        returned instead.

        Keyword Args:
            as_string (:obj:`bool`): returns data types as string
                representations, if ``True``.

        Returns:
            :obj:`dict`: A map of :obj:`str` column name to :obj:`str` or
                :obj:`type`, depending on the value of ``as_string`` kwarg.
        '''
        if as_string:
            return {item[0]: item[1] for item in self._view.schema().items()}

        return {item[0]: _str_to_pythontype(item[1]) for item in self._view.schema().items()}

    def computed_schema(self, as_string=False):
        if as_string:
            return {item[0]: item[1] for item in self._view.computed_schema().items()}

        return {item[0]: _str_to_pythontype(item[1]) for item in self._view.computed_schema().items()}

    def on_update(self, callback, mode=None):
        '''Add a callback to be fired when :func:`perspective.Table.update()` is
        called on the parent :class:`~perspective.Table`.

        Multiple callbacks can be set through calling ``on_update`` multiple
        times, and will be called in the order they are set. Callback must be a
        callable function that takes exactly 1 or 2 parameters, depending on
        whether `on_update` is called with `mode="rows"`. The first parameter is
        always `port_id`, an :obj:`int` that indicates which input port the
        update comes from. A `RuntimeError` will be thrown if the callback
        has mis-configured parameters.

        Args:
            callback (:obj:`callable`): a callable function reference that will
                be called when :func:`perspective.Table.update()` is called.
            mode (:obj:`str`): if set to "rows", the callback will be passed
                an Arrow-serialized dataset of the rows that were updated.
                Defaults to "none".

        Examples:
            >>> def updater(port_id):
            ...     print("Update fired on port", port_id)
            >>> view.on_update(updater)
            >>> table.update({"a": [1]})'
            >>> Update fired on port 0
        '''
        self._table._state_manager.call_process(self._table._table.get_id())
        mode = mode or "none"

        if not callable(callback):
            raise ValueError('Invalid callback - must be a callable function')

        if mode not in ["none", "cell", "row"]:
            raise ValueError('Invalid update mode {} - valid on_update modes are "none", "cell", or "row"'.format(mode))

        if mode == "cell" or mode == "row":
            if not self._view._get_deltas_enabled():
                self._view._set_deltas_enabled(True)

        wrapped_callback = partial(
            self._wrapped_on_update_callback, mode=mode, callback=callback)

        self._callbacks.add_callback({
            "name": self._name,
            "orig_callback": callback,
            "callback": wrapped_callback
        })

    def remove_update(self, callback):
        '''Given a callback function, remove it from the list of callbacks.

        Args:
            callback (:obj:`func`): a function reference that will be removed.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view2 = table.view()
            >>> def callback():
            ...     print("called!")
            >>> view.on_update(callback)
            >>> view2.on_update(callback)
            >>> table.update(new_data)
            called!
            >>> view2.remove_update(callback)
            >>> table.update(new_data) # callback removed and will not fire
        '''
        self._table._state_manager.call_process(self._table._table.get_id())
        if not callable(callback):
            return ValueError("remove_update callback should be a callable function!")
        self._callbacks.remove_callbacks(lambda cb: cb["orig_callback"] != callback)

    def on_delete(self, callback):
        '''Set a callback to be run when the :func:`perspective.View.delete()`
        method is called on this :class:`~perspective.View`.

        Args:
            callback (:obj:`callable`): A callback to run after
                :func:`perspective.View.delete()` method has been called.

        Examples:
            >>> def deleter():
            >>>     print("Delete called!")
            >>> view.on_delete(deleter)
            >>> view.delete()
            >>> Delete called!
        '''
        if not callable(callback):
            return ValueError("on_delete callback must be a callable function!")
        self._delete_callbacks.add_callback(callback)

    def delete(self):
        '''Delete the :class:`~perspective.View` and clean up all associated
        callbacks.

        This method must be called to clean up resources used by the
        :class:`~perspective.View`, as it will last for the lifetime of the
        underlying :class:`~perspective.Table` otherwise.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view.delete()
        '''
        self._table._state_manager.remove_process(self._table._table.get_id())
        self._table._views.pop(self._table._views.index(self._name))
        # remove the callbacks associated with this view
        self._callbacks.remove_callbacks(lambda cb: cb["name"] != self._name)
        [cb() for cb in self._delete_callbacks.get_callbacks()]

    def remove_delete(self, callback):
        '''Remove the delete callback associated with this
        :class:`~perspective.View`.

        Args:
            callback (:obj:`callable`): A reference to a callable function that
                will be removed from delete callbacks.

        Examples:
            >>> table = perspective.Table(data)
            >>> view = table.view()
            >>> view2 = table.view()
            >>> def callback():
            ...     print("called!")
            >>> view.on_delete(callback)
            >>> view2.on_delete(callback)
            >>> view.delete()
            called!
            >>> view2.remove_delete(callback)
            >>> view2.delete() # callback removed and will not fire
        '''
        if not callable(callback):
            return ValueError("remove_delete callback should be a callable function!")
        self._delete_callbacks.remove_callbacks(lambda cb: cb != callback)

    def to_arrow(self, **kwargs):
        options = _parse_format_options(self, kwargs)
        if self._sides == 0:
            return to_arrow_zero(self._view, options["start_row"], options["end_row"], options["start_col"], options["end_col"])
        elif self._sides == 1:
            return to_arrow_one(self._view, options["start_row"], options["end_row"], options["start_col"], options["end_col"])
        else:
            return to_arrow_two(self._view, options["start_row"], options["end_row"], options["start_col"], options["end_col"])

    def to_records(self, **kwargs):
        '''Serialize the :class:`~perspective.View`'s dataset into a :obj:`list`
        of :obj:`dict` containing each row.

        By default, the entire dataset is returned, though this can be windowed
        via ``kwargs``.  When ``row_pivots`` are applied, a ``__ROW_PATH__``
        column name will be generated in addition to the applied ``columns``.
        When ``column_pivots`` are applied, column names will be qualified
        with their column group name.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`list` of :obj:`dict`: A list of :obj:`dict`, where each dict
                represents a row of the current state of the
                :class:`~perspective.View`.
        '''
        return to_format(kwargs, self, 'records')

    def to_dict(self, **options):
        '''Serialize the :class:`~perspective.View`'s dataset into a :obj:`dict`
        of :obj:`str` keys and :obj:`list` values.  Each key is a column name,
        and the associated value is the column's data packed into a :obj:`list`.
        If the :class:`~perspective.View` is aggregated, the aggregated dataset
        will be returned.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`dict`: A dictionary with string keys and list values, where
                key = column name and value = column values.
        '''
        return to_format(options, self, 'dict')

    def to_numpy(self, **options):
        '''Serialize the view's dataset into a :obj:`dict` of :obj:`str` keys
        and :class:`numpy.array` values.  Each key is a column name, and the
        associated value is the column's data packed into a numpy array.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).

        Returns:
            :obj:`dict` of :class:`numpy.array`: A dictionary with string keys
                and numpy array values, where key = column name and
                value = column values.
        '''
        return to_format(options, self, 'numpy')

    def to_df(self, **options):
        '''Serialize the view's dataset into a pandas dataframe.

        If the view is aggregated, the aggregated dataset will be returned.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to ``False``).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to ``False``).


        Returns:
            :class:`pandas.DataFrame`: A DataFrame serialization of the current
                state of this :class:`~perspective.View`.
        '''
        cols = self.to_numpy(**options)
        return pandas.DataFrame(cols)

    def to_csv(self, **options):
        '''Serialize the :class:`~perspective.View`'s dataset into a CSV string.

        Keyword Args:
            start_row (:obj:`int`): (Defaults to 0).
            end_row (:obj:`int`): (Defaults to
                :func:`perspective.View.num_rows()`).
            start_col (:obj:`int`): (Defaults to 0).
            end_col (:obj:`int`): (Defaults to
                :func:`perspective.View.num_columns()`).
            index (:obj:`bool`): Whether to return an implicit pkey for each
                row (Defaults to False).
            leaves_only (:obj:`bool`): Whether to return only the data at the
                end of the tree (Defaults to False).
            date_format (:obj:`str`): How ``datetime`` objects should be
                formatted in the CSV.

        Returns:
            :obj:`str`: A CSV-formatted string containing the serialized data.
        '''
        return self.to_df(**options).to_csv(date_format=options.pop("date_format", "%Y/%m/%d %H:%M:%S"), line_terminator='\r\n' if os.name == 'nt' else '\n')

    @wraps(to_records)
    def to_json(self, **options):
        return self.to_records(**options)

    @wraps(to_dict)
    def to_columns(self, **options):
        return self.to_dict(**options)

    def _get_step_delta(self):
        pass

    def _get_row_delta(self):
        if self._sides == 0:
            return get_row_delta_zero(self._view)
        elif self._sides == 1:
            return get_row_delta_one(self._view)
        else:
            return get_row_delta_two(self._view)

    def _num_hidden_cols(self):
        '''Returns the number of columns that are sorted but not shown.'''
        hidden = 0
        columns = self._config.get_columns()
        for sort in self._config.get_sort():
            if sort[0] not in columns:
                hidden += 1
        return hidden

    def _wrapped_on_update_callback(self, **kwargs):
        '''Provide the user-defined callback function with additional metadata
        from the view.
        '''
        mode = kwargs["mode"]
        port_id = kwargs["port_id"]
        cache = kwargs["cache"]
        callback = kwargs["callback"]

        if cache.get(port_id) is None:
            cache[port_id] = {}

        if mode == "cell":
            if cache[port_id].get("step_delta") is None:
                raise NotImplementedError("not implemented get_step_delta")
            callback(port_id, cache["step_delta"])
        elif mode == "row":
            if cache[port_id].get("row_delta") is None:
                cache["row_delta"] = self._get_row_delta()
            callback(port_id, cache["row_delta"])
        else:
            callback(port_id)



# *****************************************************************************
#
# Copyright (c) 2019, the Perspective Authors.
#
# This file is part of the Perspective library, distributed under the terms of
# the Apache License 2.0.  The full license can be found in the LICENSE file.
#
import numpy as np
from datetime import date, datetime
from perspective.table import Table


class TestViewComputed(object):

    def test_view_computed_create(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(computed_columns=[{
            "column": "computed",
            "computed_function_name": "+",
            "inputs": ["a", "b"]
        }])
        assert view.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed": [6, 8, 10, 12]
        }

    def test_view_computed_multiple_dependents(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(computed_columns=[{
                "column": "computed",
                "computed_function_name": "+",
                "inputs": ["a", "b"]
            },
            {
                "column": "final",
                "computed_function_name": "pow2",
                "inputs": ["computed"]
            }
        ])
        assert view.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed": [6, 8, 10, 12],
            "final": [36, 64, 100, 144]
        }

    def test_view_computed_multiple_views_should_not_conflate(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })

        view = table.view(computed_columns=[{
            "column": "computed",
            "computed_function_name": "+",
            "inputs": ["a", "b"]
        }])

        view2 = table.view(computed_columns=[{
            "column": "computed2",
            "computed_function_name": "-",
            "inputs": ["a", "b"]
        }])

        assert view.schema() == {
            "a": int,
            "b": int,
            "computed": float
        }

        assert view2.schema() == {
            "a": int,
            "b": int,
            "computed2": float
        }

        assert view.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed": [6, 8, 10, 12]
        }

        assert view2.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed2": [-4, -4, -4, -4]
        }

    def test_view_computed_delete_and_create(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })

        view = table.view(computed_columns=[{
            "column": "computed",
            "computed_function_name": "+",
            "inputs": ["a", "b"]
        }])

        assert view.schema() == {
            "a": int,
            "b": int,
            "computed": float
        }

        assert view.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed": [6, 8, 10, 12]
        }

        view.delete()

        view2 = table.view(computed_columns=[{
            "column": "computed2",
            "computed_function_name": "-",
            "inputs": ["a", "b"]
        }])

        assert view2.schema() == {
            "a": int,
            "b": int,
            "computed2": float
        }

        assert view2.to_columns() == {
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8],
            "computed2": [-4, -4, -4, -4]
        }

    def test_view_computed_delete_and_create_with_updates(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })

        view = table.view(computed_columns=[{
            "column": "computed",
            "computed_function_name": "+",
            "inputs": ["a", "b"]
        }])

        assert view.schema() == {
            "a": int,
            "b": int,
            "computed": float
        }

        table.update({
            "a": [5, 6],
            "b": [9, 10]
        })

        assert view.to_columns() == {
            "a": [1, 2, 3, 4, 5, 6],
            "b": [5, 6, 7, 8, 9, 10],
            "computed": [6, 8, 10, 12, 14, 16]
        }

        view.delete()

        view2 = table.view(computed_columns=[{
            "column": "computed2",
            "computed_function_name": "-",
            "inputs": ["a", "b"]
        }])

        assert view2.schema() == {
            "a": int,
            "b": int,
            "computed2": float
        }

        table.update({
            "a": [5, 6],
            "b": [9, 10]
        })

        table.update({
            "a": [5, 6],
            "b": [9, 10]
        })

        assert view2.to_columns() == {
            "a": [1, 2, 3, 4, 5, 6, 5, 6, 5, 6],
            "b": [5, 6, 7, 8, 9, 10, 9, 10, 9, 10],
            "computed2": [-4, -4, -4, -4, -4, -4, -4, -4, -4, -4]
        }

    def test_view_computed_with_custom_columns(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(
            columns=["computed", "b"],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "+",
                "inputs": ["a", "b"]
            }]
        )
        assert view.to_columns() == {
            "b": [5, 6, 7, 8],
            "computed": [6, 8, 10, 12]
        }

    def test_view_computed_with_row_pivots(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(
            row_pivots=["computed"],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "+",
                "inputs": ["a", "b"]
            }]
        )
        assert view.to_columns() == {
            "__ROW_PATH__": [[], ["6"], ["8"], ["10"], ["12"]],
            "a": [10, 1, 2, 3, 4],
            "b": [26, 5, 6, 7, 8],
            "computed": [36.0, 6.0, 8.0, 10.0, 12.0]
        }
    
    def test_view_computed_with_column_pivots(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(
            column_pivots=["computed"],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "+",
                "inputs": ["a", "b"]
            }]
        )
        assert view.to_columns() == {
            "6|a": [1, None, None, None],
            "6|b": [5, None, None, None],
            "6|computed": [6, None, None, None],
            "8|a": [None, 2, None, None],
            "8|b": [None, 6, None, None],
            "8|computed": [None, 8, None, None],
            "10|a": [None, None, 3, None],
            "10|b": [None, None, 7, None],
            "10|computed": [None, None, 10.0, None],
            "12|a": [None, None, None, 4],
            "12|b": [None, None, None, 8],
            "12|computed": [None, None, None, 12.0]
        }

    def test_view_computed_with_row_column_pivots(self):
        table = Table({
            "a": [1, 2, 3, 4],
            "b": [5, 6, 7, 8]
        })
        view = table.view(
            column_pivots=["computed"],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "+",
                "inputs": ["a", "b"]
            }]
        )
        assert view.to_columns() == {
            "6|a": [1, None, None, None],
            "6|b": [5, None, None, None],
            "6|computed": [6.0, None, None, None],
            "8|a": [None, 2, None, None],
            "8|b": [None, 6, None, None],
            "8|computed": [None, 8.0, None, None],
            "10|a": [None, None, 3, None],
            "10|b": [None, None, 7, None],
            "10|computed": [None, None, 10.0, None],
            "12|a": [None, None, None, 4],
            "12|b": [None, None, None, 8],
            "12|computed": [None, None, None, 12.0]
        }
    
    def test_view_computed_with_sort(self):
        table = Table({
            "a": ["a", "ab", "abc", "abcd"]
        })
        view = table.view(
            sort=[["computed", "desc"]],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "length",
                "inputs": ["a"]
            }]
        )

        assert view.to_columns() == {
            "a": ["abcd", "abc", "ab", "a"],
            "computed": [4, 3, 2, 1]
        }

    def test_view_computed_with_filter(self):
        table = Table({
            "a": ["a", "ab", "abc", "abcd"]
        })
        view = table.view(
            filter=[["computed", ">=", 3]],
            computed_columns=[{
                "column": "computed",
                "computed_function_name": "length",
                "inputs": ["a"]
            }]
        )

        assert view.to_columns() == {
            "a": ["abc", "abcd"],
            "computed": [3, 4]
        }

    def test_view_day_of_week_date(self):
        table = Table({
            "a": [date(2020, 3, i) for i in range(9, 14)]
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "day_of_week",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": str
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 3, i) for i in range(9, 14)],
            "bucket": ["2 Monday", "3 Tuesday", "4 Wednesday", "5 Thursday", "6 Friday"]
        }

    def test_view_day_of_week_datetime(self):
        table = Table({
            "a": [datetime(2020, 3, i, 12, 30) for i in range(9, 14)]
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "day_of_week",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": datetime,
            "bucket": str
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 3, i, 12, 30) for i in range(9, 14)],
            "bucket": ["2 Monday", "3 Tuesday", "4 Wednesday", "5 Thursday", "6 Friday"]
        }

    def test_view_month_of_year_date(self):
        table = Table({
            "a": [date(2020, i, 15) for i in range(1, 13)]
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_of_year",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": str
        }
        assert view.to_columns() == {
            "a": [datetime(2020, i, 15) for i in range(1, 13)],
            "bucket": ["01 January", "02 February", "03 March", "04 April", "05 May", "06 June", "07 July", "08 August", "09 September", "10 October", "11 November", "12 December"]
        }

    def test_view_month_of_year_datetime(self):
        table = Table({
            "a": [datetime(2020, i, 15) for i in range(1, 13)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_of_year",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": datetime,
            "bucket": str
        }
        assert view.to_columns() == {
            "a": [datetime(2020, i, 15) for i in range(1, 13)],
            "bucket": ["01 January", "02 February", "03 March", "04 April", "05 May", "06 June", "07 July", "08 August", "09 September", "10 October", "11 November", "12 December"]
        }

    # bucketing
    def test_view_day_bucket_date(self):
        table = Table({
            "a": [date(2020, 1, 1), date(2020, 1, 1), date(2020, 2, 29), date(2020, 3, 1)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "day_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), datetime(2020, 1, 1), datetime(2020, 2, 29), datetime(2020, 3, 1)],
            "bucket": [datetime(2020, 1, 1), datetime(2020, 1, 1), datetime(2020, 2, 29), datetime(2020, 3, 1)]
        }

    def test_view_day_bucket_date_with_null(self):
        table = Table({
            "a": [date(2020, 1, 1), None, date(2020, 2, 29), date(2020, 3, 15)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "day_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), None, datetime(2020, 2, 29), datetime(2020, 3, 15)],
            "bucket": [datetime(2020, 1, 1), None, datetime(2020, 2, 29), datetime(2020, 3, 15)]
        }

    def test_view_day_bucket_datetime(self):
        table = Table({
            "a": [datetime(2020, 1, 1, 5), datetime(2020, 1, 1, 23), datetime(2020, 2, 29, 1), datetime(2020, 3, 1, 0)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "day_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": datetime,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1, 5), datetime(2020, 1, 1, 23), datetime(2020, 2, 29, 1), datetime(2020, 3, 1, 0)],
            "bucket": [datetime(2020, 1, 1), datetime(2020, 1, 1), datetime(2020, 2, 29), datetime(2020, 3, 1)]
        }

    def test_view_month_bucket_date(self):
        table = Table({
            "a": [date(2020, 1, 1), date(2020, 1, 28), date(2020, 2, 29), date(2020, 3, 15)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), datetime(2020, 1, 28), datetime(2020, 2, 29), datetime(2020, 3, 15)],
            "bucket": [datetime(2020, 1, 1), datetime(2020, 1, 1), datetime(2020, 2, 1), datetime(2020, 3, 1)]
        }

    def test_view_month_bucket_date_with_null(self):
        table = Table({
            "a": [date(2020, 1, 1), None, date(2020, 2, 29), date(2020, 3, 15)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": date,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), None, datetime(2020, 2, 29), datetime(2020, 3, 15)],
            "bucket": [datetime(2020, 1, 1), None, datetime(2020, 2, 1), datetime(2020, 3, 1)]
        }

    def test_view_month_bucket_datetime(self):
        table = Table({
            "a": [datetime(2020, 1, 1), datetime(2020, 1, 28), datetime(2020, 2, 29), datetime(2020, 3, 15)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": datetime,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), datetime(2020, 1, 28), datetime(2020, 2, 29), datetime(2020, 3, 15)],
            "bucket": [datetime(2020, 1, 1), datetime(2020, 1, 1), datetime(2020, 2, 1), datetime(2020, 3, 1)]
        }

    def test_view_month_bucket_datetime_with_null(self):
        table = Table({
            "a": [datetime(2020, 1, 1), None, None, datetime(2020, 3, 15)],
        })
        view = table.view(
            computed_columns=[
                {
                    "column": "bucket",
                    "computed_function_name": "month_bucket",
                    "inputs": ["a"]
                }
            ]
        )
        assert view.schema() == {
            "a": datetime,
            "bucket": date
        }
        assert view.to_columns() == {
            "a": [datetime(2020, 1, 1), None, None, datetime(2020, 3, 15)],
            "bucket": [datetime(2020, 1, 1), None, None, datetime(2020, 3, 1)]
        }
